// primitiv.cxx 
// W. Langdon cs.ucl.ac.uk
// version "$Revision: 1.18 $"

// Copyright Andy Singleton, 1993,1994
// This code is released for non-commercial use only
// For questions or upgrades contact:
// Andy Singleton, Creation Mechanics Inc.
// PO Box 248, Peterborough, NH 03458
// Internet: p00396@psilink.com
// Compuserve 73313,757
// Phone: (603) 563-7757
 
//Modifications (in reverse order)
//WBL 30 Apr 1999 Conversion for SiliconGraphics cc -c -64 -mips4

//WBL  9 Jan 1998  add SMC_GP (based on RMP's smc_gp)

//WBL  8 Mar 1998  Restore Not, And, Or, If

//WBL 30 Nov 1995  Add order reversed SUB and DIV

//WBL 26 Oct 1995  Restore Mul, Div and Abs

//WBL 15 Oct 1995  Add TIMINGS

//WBL 23 Jan 1995  Restore IFlteEval

//WBL  2 May 1994  Real number support removed

//**************************************************************************
//  Standard Functions
//	includes ConstFunc (numeric constants)
//		ADD,SUB,MUL,DIV,SINE,ABS,ADD4,PROG4,BOOL,NOT,AND,OR,IF,IFlte
//**************************************************************************

#include "pch.h"
#include "chrome.h"
#include "prob.h"
#ifdef TIMINGS
#include "time.h"
#endif
#include "primitiv.h"

extern char scratch[100];
#ifdef FASTEVAL
	extern evalnode* IpGlobal;
	extern Chrome* ChromeGlobal;
#endif

//#ifdef SMC_GP
//OPDEF(ConstEval) {assert(1==0); return 0;}
//#else
OPDEF(ConstEval) {return GETVAL;}
//#endif /*SMC_GP*/

#ifndef ANT_LIB
				// Add two arguments.  Note that argnum = 2
OPDEF(AddEval) {return EVAL + EVAL;}

				// Subtract two arguments
OPDEF(SubEval) {return EVAL - EVAL;}
OPDEF(SubREval) {return -EVAL + EVAL;}  //reversed order of arguments

//OPDEF(MulEval) {retval rval= EVAL * EVAL; return BOUNDF(rval);}
OPDEF(MulEval) {return EVAL * EVAL;} //retval!=float

				// "Protected" division
OPDEF(DivEval)
{
	retval numerator,denominator;//,rval;
	numerator = EVAL;
	denominator = EVAL;
	if (denominator !=0)
//		rval = numerator/denominator;
		return numerator/denominator;	//retval!=float
	else
//		rval=1;
		return 1;			//retval!=float
//	return BOUNDF(rval);
}

OPDEF(DivREval) //reversed order of arguments
{
	retval numerator,denominator;//,rval;
	denominator = EVAL;
	numerator = EVAL;
	if (denominator !=0)
//		rval = numerator/denominator;
		return numerator/denominator;	//retval!=float
	else
//		rval=1;
		return 1;			//retval!=float
//	return BOUNDF(rval);
}

OPDEF(AbsEval)			// Absolute value
{
	retval arg = EVAL;
        return (arg > 0 ? arg : -arg);
}

OPDEF(IfEval)           // Conditional. Thrown in for reference.
					// If(condition>0, dothis, otherwise dothat)
					// Conditionals, loops, subroutines come from
					// manipulating the instruction pointer
{
#ifdef SMC_GP
	const retval vif = EVAL;
	const retval left = EVAL;
	const retval right = EVAL;
	return (vif & left) | ((~vif) & right); /* Bitwise IF */
#else
	retval rval;
	if (EVAL>0)
	{
		rval=EVAL;
		IP++;           // Jump the third expression
		TRAVERSE();
		IP--;           // And back up one (since EVAL increments)
	}
	else {
		IP++;
		TRAVERSE();    // Jump the second expression
		IP--;                                   // Back up for EVAL
		rval=EVAL;
	}
	return rval;
#endif
}

OPDEF(IflteEval)	//IfLTE(condition1islessthan,condition2,dothis,dothat)
{
	retval rval;
        rval=EVAL;
	if (rval<=EVAL)
	{
#ifdef TIMINGS
		ThisTimmer->opt_time(IP);
#endif
		rval=EVAL;
		IP++;           // Jump the third expression
		TRAVERSE();
		IP--;           // And back up one (since EVAL increments)
	}
	else {
		IP++;
		TRAVERSE();    // Jump the second expression
		IP--;                                   // Back up for EVAL
#ifdef TIMINGS
		ThisTimmer->opt_time(IP);
#endif
		rval=EVAL;
	}
	return rval;
}
#endif //ANT_LIB
OPDEF(NotEval)		// NOT(arg)
{
#ifdef SMC_GP
	return ~(EVAL);   /* Bitwise NOT */
#else
	return (EVAL > 0 ? 0 : 1);
#endif
}

// watch out here - your compiler may optimize away an EVAL
OPDEF(AndEval)		// AND(arg1,arg2)
{
#ifdef FASTTRAVERSE
	if(EVAL <= 0) {
	  IP++;
	  TRAVERSE();    // Jump the second expression
	  IP--;          // Back up for EVAL
	  return 0;
	}
	return EVAL;
#else
#ifdef SMC_GP
	retval arg1 = EVAL; //cc.64
	retval arg2 = EVAL;
	return arg1 & arg2;   /* Bitwise AND */
#else
	retval arg1 = EVAL;
	retval arg2 = EVAL;
	return (arg1 > 0 && arg2 > 0 ? 1 : 0);
#endif
#endif /*FASTTRAVERSE*/
}

OPDEF(OrEval)		// OR(arg1,arg2)
{
#ifdef FASTTRAVERSE
	if(EVAL > 0) {
	  IP++;
	  TRAVERSE();    // Jump the second expression
	  IP--;          // Back up for EVAL
	  return 1;
	}
	return EVAL;
#else
#ifdef SMC_GP
	retval arg1 = EVAL;//cc.64
	retval arg2 = EVAL;
	return arg1 | arg2;   /* Bitwise OR  */
#else
	retval arg1 = EVAL;
	retval arg2 = EVAL;
	return (arg1 > 0 || arg2 > 0 ? 1 : 0);
#endif
#endif /*FASTTRAVERSE*/
}
