// stats.ccc  version to support queue2.cc, from GPQUICK-2.1 chrome.cc
// W.Langdon cs.ucl.ac.uk $Revision: 1.89 $

//Modifications (reverse order):
// WBL 12 Jul 2016  Add exon
// WBL  5 Jul 2016  Add DisplayZStats

// WBL 23 Jun 2016  Add nmax

// WBL  7 Sep 1999  Add vhits, vfit

// WBL    Aug 1999  Add DisplaySStats

// WBL 29 Jun 1999  Add Homologous diference monitoring code

// WBL  4 Jun 1999  Add rawfitness

// WBL 30 Apr 1999  For SiliconGraphics cc -c -64 -mips4, 
//                  surpress unused stats memory and NTests (not POLY_LIB)

// WBL 20 Nov 1998  Add display of DepthMax to DisplayStats

// WBL  8 Aug 1998  Move costats to chrome.h

// WBL 31 Jul 1998  Allow funccount to be bigger than 32 (but only display 
//                  stats on first 32) opcodes. Maybe some problems with about
//                  32 functions?

// WBL  7 Jan 1998  move class stats into chrome.h

// WBL  6 Jan 1998  Add explicit display of zero fitter neighbours

// WBL 31 Dec 1997  convert hashtable and hashchain into subclasses
//                  Add stats for all.cc

// WBL 13 Aug 1997  Make DisplayCStats robust against divide by zero errors.

// WBL  8 May 1997  Add first level of support for PDGP

// WBL 12 Feb 1997  GatherPStats so includes depths>=MAXTREEDEPTH
//                  DisplayPStats and DisplayLPStats to display it
//                  Doubtless output format of DisplayPStats could be better

// WBL 16 Feb 1997  Add costats

// WBL  7 Feb 1997  Add array bound checks (MAXTREEDEPTH isnt bigenought)

// WBL  1 Jan 1997  BUGFIX complete change of meaning implemented on 4-Apr-96
//                  so non-integer scores now work as they should
//                  Ensure nfit has same meaning as in integer scores case

// WBL  1 Jan 1997  Get DisplayDStats working for tournament sizes other than 4

// WBL 31 Dec 1996  BUGFIX avoid hashchain::~hashchain() calling itself

// WBL 31 Dec 1996  BUGFIX remove delete[] table from hashtable::include()
//                  Improve ERROR display in Pop::DisplayPStats

// WBL 16 Dec 1996  Make GatherPStats gather opcode data by tree depth

// WBL  4 Apr 1996  Make DisplayDStats assume integer scores and use blickle
//                  formula. NB covariance of nrank and nfit changed. 
//                  nrank nlonger supported.

// WBL 16 Mar 1996  Make DisplayDupStats record and display detailed stats

// WBL 15 Mar 1996  Add hashcode, Chrome::same and DisplayDupStats

// WBL 26 Feb 1996  Restore version for stack (from chrome.cxx r1.4) and 
//                  splice in using ifndef STACK_LIB (ie not gp.h)

//WBL 22 Sep 95  Replace list.h by prob.h

// WBL 23 May 1995  Add DisplayCStats

// WBL 12 May 1995  Add gathering problem dependant data

// WBL  8 May 1995  Output from DisplayLPStats >10% and make easier to find

// WBL  7 May 1995  Split GatherPStats from DisplayPStats and add 
//                  DisplayLPStats

// WBL  1 May 1995  Add DisplayPStats

// WBL 23 Apr 1995  Add Ntests_run

// WBL 20 Apr 1995  Add TIMINGS

// WBL 19 Feb 1995  Remove unneccessary output (to reduce disk usage)

// WBL 16 Feb 1995  Allow program scores to be of type scoretype

// WBL 12 Feb 1995  OK see if can get working now

// WBL 23 Jan 1995  Partial conversion to problem independance
//                  add gp.h, Primitives
//                  HACK to make COMPILE

// WBL 22 Nov 1994  Refit lost change.
//                  Change Badadf1 to adf1ver

// WBL 13 Sep 1994  Add displaying adf1

// WBL 13 Sep 1994  Make compilable by solaris gcc

// WBL 25 Aug 1994  Remove BestFitness check

// WBL 22 Aug 1994  Add queue2.h
//                  Take hits value from nfitness, rather than Chrome
//                  Add display of mem_used.

// WBL 15 Aug 1994  Add display of hits and total_hits

// WBL	6 Jul 1994  Fix bug whereby may not set last few values of nfit

// WBL  1 Jul 1994  New file based upon chrome.cxx Revision: 1.9
//                  Add pTraceDynamic.
//		    Display primative frequency, covarience and correllation

// WBL 28 Jun 1994  Make nfit take average of cube rather than linear average
//                  Tidy displays in DisplayDStats

// WBL 27 Jun 1994  Correct display of gencount in DisplayStats
//                  Add class stats and make DisplayStats use it.
//                  Complete impementation of DisplayDStats

// WBL 24 Jun 1994  Direct crossover to a particular tree
//                  Protect sqrt from rounding errors

// WBL 14 Jun 1994  Add support for pPopSeed, pTestSeed and PTrace

// WBL 14 Jun 1994  Add ChromeParams::Save() and ChromeParams::Load()
//                  based upon gpcplus3.0 cpv.cc.
//                  Add Pop::DisplayDStats()
//		    Add Pop::Write()

// WBL 16 May 1994  Add Pop::DisplayStats(). Fix MULTREE CrossOver bug

// WBL 12 May 1994  fix % by zero error in rnd() on SETNODE if varnum==0,
//		    Probably only arises if you have the wrong compiler switch

// WBL  5 May 1994  Make pMaxAge = 0 disable age limit,
//		    Add support for P-M random numbers

// GPQUICK
// C++ prefix stack implementation
// Divides GP system into classes:

#include <math.h>	//for DisplayStats
#include <assert.h>
#include "pch.h"
#include "chrome.h"
#include "prob.h"
#ifndef STACK_LIB
//#include "Primitives.h"
#ifdef QUEUE_LIB
#include "queue2.h"
#else
#include "gp.h"
#endif /*QUEUE_LIB*/
#endif

double stats::Variance()					//WBL
{ 	double v = (sqr - double(cases*double(Mean())*double(Mean())) )/cases;
	if ( v >= 0 ) return v; else return 0.0; //protect against rounding
}

void stats::include( float in )  			//WBL
{	cases++;
	sum += in;
    	sqr += double(in) * double(in);
	if ( max < in ) {max = in; nmax = 1;}
	else if ( max == in ) nmax++;
	if ( min > in ) min = in;
}//end stats::include


//NB covar stats assume primitive was present or wasnt cant use this code
//for covariance of continous things like length
void stats::display ( char* name, ostream & fout,
                      stats* own_stats, stats* co_stats,
                      costats* covar)
                                                      //WBL
{
if ( cases > 0 )
    {	float mean     = Mean();
	double variance = Variance();
	fout << "Mean " << name << "= " << mean;
//	fout << "\tvar = " << variance;
	fout << "\tSD = "<<sqrt(variance);
	if ( covar != NULL)
	     { 	fout << "\tcovar " << covar->Covariance();
		fout << "\tcorel " << covar->Correlation();
	     };
	if ( own_stats != NULL && co_stats != NULL)
	     { 	fout << "\tcovar ";
//
//covariance = 	sum (x(i)*y(i))
//              ---------------  -  mean x * mean y
//               num cases
//
//In our case sum (x(i)*y(i)) = sum because we count y times. 
//The mean x is number of x (ie cases) divided by the popsize
//
		float covar;

		covar = (float)sum - (float) cases * co_stats->Mean();
		covar = covar/(float) own_stats->Cases();
		fout << covar;

// Correlleation co-efficient =             covariance
//				------------------------------------
//				sqrt( own variance * others variance
//
// Own variance includes cases where chrome contains none of the
// opcode, nb including those not in cases
//
// own variance = sqr / popsize - ( sum / popsize )**2
//
		{
		double d = sqrt( own_stats->Variance() * co_stats->Variance());
		if ( d > 0 )			//protect rounding error
		      { fout << "\tcorel ";
			fout << covar / d;
		      }
		} 
	     }
	fout << "\tmin = " << min;
	fout << "\tmax = " << max << " " << nmax <<endl;
    }//endif
}//end stats::display

void Pop::DisplayStats(ostream & fout) const //WBL
{
	time_t now;
//	ChromeParams* params;
//	Problem* problem;
	time (&now);

	fout << ctime(&now);
#ifdef STACK_LIB
	fout << "popsize =\t"<< popsize  <<endl; 
	fout << "individuals =\t" <<gencount+1  <<endl;
	fout << "BestFitness =\t" <<*BestFitness  <<endl;
	fout << "BestMember =\t" <<BestMember  <<endl;
	fout << "initeval =\t" <<initeval  <<endl;
	fout << "nexteval =\t" <<nexteval  <<endl;
#endif
	stats fitness;
#ifdef RAWFITSTAT
	stats rawfitness;
#endif /*RAWFITSTAT*/
	stats length;
	stats depth;
#ifdef INTRON64
	stats exon;
	stats nsol;
#endif /*INTRON64*/
#ifdef PARETO
	stats hits [last_op+1];
	stats total_hits;
#endif
//ifdef BRACK_LIB || CALC_LIB || GRID_LIB || KARP_LIB || LIST_LIB || QUEUE_LIB || STACK_LIB
//	stats memory;
/*endif*/
#ifdef MINC_ADF
	stats adf1;
#endif
#ifdef TIMINGS
	stats cpu;
#endif
#ifdef POLY_LIB
	stats Ntests;
#endif
#ifdef ALLFUNC_LIB
	stats Ntests;
#endif
	stats vfit;
	stats vhits;
	for (UINT i=0;i<popsize;i++)
	      { fitness.include (pop[i]->nfitness->fvalue);
#ifdef PDGP
		if(!params->params[pPDGP]) 
#endif /*PDGP*/
	       const int size=pop[i]->tree_starts[NUM_TREES-1] +
                              pop[i]->SubLen(pop[i]->tree_starts[NUM_TREES-1]);
	       length.include(size);
#ifdef INTRON64
	       exon.include(pop[i]->nfitness->exon);
	       nsol.include(pop[i]->nfitness->nsol);
#endif /*INTRON64*/
	       depth.include(pop[i]->DepthMax(pop[i]->tree_starts[NUM_TREES-1],
					      size-1));
#ifdef PARETO
		scoretype total = 0;
#endif
#ifndef STACK_LIB
                                                             //force right type
		ptrmyfitnessvalue temp = ptrmyfitnessvalue (pop[i]->nfitness);
#ifdef RAWFITSTAT
	      rawfitness.include(temp->rawfit);
#endif /*RAWFITSTAT*/
#ifdef PARETO
		for (int j=0; j<=last_op; j++)
			{ total += temp->hits[j];
			  hits[j].include (temp->hits[j]);
		        }
		total_hits.include (total);
#endif
		//memory.include (temp->mem_used);
#ifdef MINC_ADF
		adf1.include(temp->adf1);
#endif
#ifdef TIMINGS
		cpu.include(temp->cpu_time);
#endif
#ifdef POLY_LIB
		Ntests.include(temp->Ntests_run);
#endif /*POLY_LIB*/
		vfit.include(temp->vfit);
		vhits.include(temp->vhits);
#endif
	      }
                                                      
	fitness.display( "fitness", fout );
#ifdef RAWFITSTAT
	rawfitness.display( "rawfitness", fout );
#endif /*RAWFITSTAT*/
#ifdef PDGP
	if(!params->params[pPDGP]) 
#endif /*PDGP*/
	length.display( "length ", fout );
	depth.display( "depth ", fout );
#ifdef INTRON64
	exon.display( "exon ", fout );
	nsol.display( "nsol ", fout );
#endif /*INTRON64*/

#ifdef PARETO
	for (int j=0; j<=last_op; j++) 
		{ problem->WriteTreeName(j, fout);
		  fout << "\t";
		  hits[j].display("", fout);
	        }
	fout << "total\t";
	total_hits.display( "", fout );
#endif
	//fout << "memory\t";
	//memory.display( "", fout );
#ifdef MINC_ADF
	fout << "adf1ver\t";
	adf1.display( "", fout );
#endif
#ifdef TIMINGS
	fout << "cpu\t";
	cpu.display( "", fout );
#endif
#ifdef POLY_LIB
	fout << "tests\t";
	Ntests.display( "", fout );
#endif /*POLY_LIB*/
	fout << "vfit\t";
	vfit.display( "", fout );
	fout << "vhits\t";
	vhits.display( "", fout );

}//end Pop::DisplayStats

void Hist_add(int hist_size[],int& max, const int size, const int odd) {
  if(odd) assert(size%2);
  else    assert(size%2==0);
  const int i = size/2;
  assert(i<EXPRLEN/2);
  hist_size[i]++;
  if(i>max) max = i;
}
#define Max(x,y) ((x)>(y))? (x) : (y)
void print_hist3(const int gencount, const char* name,
		 const int A[], const int maxA, const int oddA, 
		 const int B[], const int maxB, const int oddB, 
		 const int C[], const int maxC, const int oddC, 
		 ostream & fout){
  const int max = Max(maxA,Max(maxB,maxC));
  for(int i=0;i<=max;i++) {
    fout<<gencount<<" "<<name;
    fout<<i*2+oddA<<" "<<A[i]<<" ,\t";
    fout<<i*2+oddB<<" "<<B[i]<<" ,\t";
    fout<<i*2+oddC<<" "<<C[i]<<endl;
  }
}

void Pop::DisplayZStats(ostream & fout) const //WBL
{
#if NUM_TREES != 1
Error DisplayZStats assumes one tree
#endif
  int hs=-1;
  int hist_size[EXPRLEN/2];     memset(hist_size,    0,(EXPRLEN/2)*sizeof(int));
  int hss=-1;
  int hist_subsize[EXPRLEN/2];  memset(hist_subsize, 0,(EXPRLEN/2)*sizeof(int));
  int hrs=-1;
  int hist_rootsize[EXPRLEN/2]; memset(hist_rootsize,0,(EXPRLEN/2)*sizeof(int));

  for (UINT i=0;i<popsize;i++) { 
    const int size = pop[i]->SubLen(0);
    Hist_add(hist_size,hs,size,1);
    for (int j=0; j<size; j++) { 
      const int subsize=pop[i]->SubLen(j);
      Hist_add(hist_subsize, hss,subsize     ,1);
      Hist_add(hist_rootsize,hrs,size-subsize,0);
    }
  } 
  print_hist3(gencount+1,"sizes ",
	      hist_size,    hs, 1,
	      hist_subsize, hss,1,
	      hist_rootsize,hrs,0,fout);
}//end Pop::DisplayZStats

ofstream* ppopdata = NULL;

void Pop::DisplaySStats() const //WBL
{
#define popdata *ppopdata
  if(ppopdata==NULL) {
    ppopdata = new ofstream(params->params_str[pPopfile]);
    if(!popdata || strlen(params->params_str[pPopfile])<=0) {//ignores nullfile
      cerr<<"ERROR bad pop file name `"<<params->params_str[pPopfile]<<"'\n";
      cout<<"ERROR bad pop file name `"<<params->params_str[pPopfile]<<"'\n";
      return;
    } }
  if(!popdata || strlen(params->params_str[pPopfile])<=0) return;

  for (UINT i=0;i<popsize;i++) { 
    popdata<<gencount+1<<"\t";
    popdata<<pop[i]->nfitness->fvalue<<"\t";
    popdata<<ptrmyfitnessvalue(pop[i]->nfitness)->Ntests_run<<"\t";
    const int size=pop[i]->tree_starts[NUM_TREES-1] +
                   pop[i]->SubLen(pop[i]->tree_starts[NUM_TREES-1]);
    popdata<<size<<"\t";
    popdata<<pop[i]->DepthMax(pop[i]->tree_starts[NUM_TREES-1],size-1)<<"\t";
    popdata<<ptrmyfitnessvalue(pop[i]->nfitness)->vfit<<"\t";
    popdata<<ptrmyfitnessvalue(pop[i]->nfitness)->vhits;
    popdata<<endl;
  }
  popdata<<endl;
//popdata.close(); delete ppopdata;
}//end Pop::DisplaySStats


#ifdef PDGP
int ChromeParams::netdepth(const int net) const
{
  if(params[pPDGP])
    return net_limits[net+1] - net_limits[net];
  else
    return (params[pMaxDepth]!=0)? params[pMaxDepth] : MAXTREEDEPTH;
}
#define MaxDepth params->netdepth(t)
#endif /*PDGP*/

#define QARGNUM(op) (problem->funclist[op]->argnum)
void Pop::DisplayDStats(ostream & fout ) const //WBL
//
// Convert chromosome nfitness into a rank, and then into an averag
// rank and then into an approximate chance of selection. Cf roulette
// wheel selection.
//    Gather and display loads of statistics by tree/opcode
//
{
//float nfit [popsize];    //equivelent roulette wheel
float* nfit = new float [popsize];    //equivelent roulette wheel

#ifndef PDGP
const int MaxDepth = (params->params[pMaxDepth]!=0)? 
                         params->params[pMaxDepth] : MAXTREEDEPTH;
assert(MaxDepth<=MAXTREEDEPTH);
#endif /*PDGP*/

#ifdef INT_SCORE
{
int score[max_fitness+1];
float nr4[max_fitness+1];
memset(score,0,sizeof(score));

for (int i = 0; i < popsize; i++ ) score[int(pop[i]->nfitness->fvalue)]++;

assert (params->params[pTournSize] == 4);

int sum = 0;
float r4 = 0;

for (i = 0; i <= max_fitness; i++ ) {
	if(score[i] != 0) {
		float old  = r4;
		sum       += score[i];
		float r    = float (sum) / float (popsize);
		r4         = r*r*r*r;
		nr4[i]     = (r4 - old)*float(popsize)/float(score[i]);
	}
}//end for i
for (UINT j = 0; j<popsize; j++) nfit[j] = nr4[int(pop[j]->nfitness->fvalue)];
}
UINT i;
#else
//Chrome* rank [popsize];
//float nrank [popsize];   //normalised rank
Chrome** rank = new Chrome* [popsize];
//float* nrank  = new float   [popsize];   //normalised rank
UINT* pop_index = new UINT [popsize];

//set rank by sorting population. Sort not efficient but good enough
{for (UINT i = 0; i < popsize; i++ ) {rank [i] = pop[i]; pop_index[i] = i;}}
UINT j;
{for (UINT i = 0; i < popsize; i++ )
      { BOOL swapflag = FALSE;
	for (j = 1; j < popsize; j++ )
	      { if ( rank[j-1]->nfitness->IsBetter(rank[j]->nfitness) )
		      { Chrome* r  = rank [j-1];
			rank [j-1] = rank [j];
			rank [j]   = r;
		        int index       = pop_index [j-1];
			pop_index [j-1] = pop_index [j];
			pop_index [j]   = index;
			swapflag   = TRUE;
                      }
	      }//end loop j
	if (swapflag == FALSE) break;
      }}//end loop i

//Nolonger true WBL 25/8/94
//assert( BestFitness->fvalue == rank[popsize-1]->nfitness->fvalue ); //check

//take large population approximation
#define fitnessofrank(rank) (params->params[pTournSize]* \
               pow(double(rank)/double(popsize),params->params[pTournSize]-1))

UINT k = 0;
double sum_normalised_fitness = fitnessofrank(1);

UINT l;
for ( j = 1; j < popsize; j++ )
      { //float J = j;
	//float K = k;
	if ( rank[j]->nfitness->IsBetter(rank[k]->nfitness))
	      { 
//		nrank [ k ] = (J*(J-1)-(K-1)*K)/(2*(J-K));
		nfit  [ pop_index[k] ] = sum_normalised_fitness/(j-k);
		for ( l = k+1; l<j; l++ )
		      { //nrank[ l ] = nrank [ k ];
			nfit [ pop_index[l] ] = nfit  [ pop_index[k] ];
		      }
		k = j;
		sum_normalised_fitness = 0.0;
	      }//end j is not same rank as k
	sum_normalised_fitness += fitnessofrank(j+1);
      }

for ( l = k; l<popsize; l++ )
	      { //float J = popsize; //NB include j
		//float K = k;
//		nrank [ l ] = (J*(J-1)-(K-1)*K)/(2*(J-K));
		nfit  [ pop_index[l] ] = sum_normalised_fitness/(popsize-k);
	      }
//delete [] nrank;
delete [] pop_index;
delete [] rank;

#endif

stats length [NUM_TREES];
costats length_nfit [NUM_TREES];
costats length_fit [NUM_TREES];
struct x
{ stats frequency; /* stats length; stats children;
  stats fitness;   stats nrank;*/  stats nfit;     }; 
x summary;
//x counts [NUM_TREES][problem->funccount];

#define funccount_max 256

x counts [NUM_TREES][funccount_max][MAXTREEDEPTH+1];

{for (UINT i = 0; i < popsize; i++ )
{
//summary.frequency JUNK
//summary.length    JUNK
//summary.children.	include( pop[i]->num_children );
//summary.fitness.	include( pop[i]->nfitness->fvalue );
//summary.nrank.		include( nrank [i] );
summary.nfit.		include( nfit [i] );

int len;
for ( int t = 0; t<NUM_TREES; t++)
      { 
//	int f [problem->funccount];
	int f [funccount_max][MAXTREEDEPTH+1];
	memset(f,0,sizeof(f));
#ifdef PDGP
	if(params->params[pPDGP]) {
	  for(int d = params->net_limits[t]; d < params->net_limits[t+1]; d++){
	    const int sp = 1+d-params->net_limits[t];
	    assert(sp<=MAXTREEDEPTH);
	    for(int y = 0; y < params->pdgpwidth[d]; y++) {
	      const int op = pop[i]->pdgp(d,y)->op;
	      if(op<funccount_max) {
		f[op][sp]++;
		counts[t][op][sp].nfit.include ( nfit[i] );
	      }
	    }
	  };//end for each op in this net
	}
	else
#endif /*PDGP*/
        { 
	len = pop[i]->SubLen(pop[i]->tree_starts[t]);
	length[t].  include ( len );
	length_nfit[t].include (len, nfit[i]);
	length_fit[t].include  (len, pop[i]->nfitness->fvalue);
        int tree_stack [EXPRLEN];
        tree_stack[0] = -1;
	int sp = 0;

	{for ( int n = pop[i]->tree_starts[t];
              n < (pop[i]->tree_starts[t] + len);
	      n++ 				)
	{
	        while (tree_stack[sp] == 0) { //remove finished branches
		        --sp; //-O3 bug in mips4 cc?
			tree_stack[sp] = tree_stack[sp] - 1;
		}
		++sp;
		const int op = pop[i]->expr[n].op;
		tree_stack[sp] = QARGNUM(op);
		if(op<funccount_max) {
		if(sp<MAXTREEDEPTH){f[op][sp]++;

//counts[t][pop[i]->expr[n].op].length.  include ( len );
//counts[t][pop[i]->expr[n].op].children.include ( pop[i]->num_children );
//counts[t][pop[i]->expr[n].op].fitness. include ( pop[i]->nfitness->fvalue );
//counts[t][rank[i]->expr[n].op].nrank.   include ( nrank[i] );
counts[t][op][sp].nfit.    include ( nfit[i] );
				  }
		else               {f[op][MAXTREEDEPTH]++;
counts[t][op][MAXTREEDEPTH].nfit.include ( nfit[i] );
				  }
	      }
	}};//end for each op in this tree
        }//endif pPDGP

	{for ( int op = 0; (op<problem->funccount)&&(op<funccount_max); op++ )
	   for ( int d = 1; d <= MaxDepth; d++ )
	      { counts[t][op][d].frequency.include ( f[op][d] ); }}

      };//end for each tree
}};//end for whole population

//summary.children.display ("tree child"  , fout );
//summary.fitness. display ("tree fitness", fout );
//summary.nrank.   display ("tree nrank"  , fout );
summary.nfit.    display ("tree nfit"   , fout );

for ( int t = 0; t<NUM_TREES; t++)
{
fout << endl;
problem->WriteTreeName(t, fout);
fout << "\t";

#ifdef PDGP
if(!params->params[pPDGP])
#endif /*PDGP*/
{ 
length[t].display ("tree size", fout, NULL, NULL, &length_nfit[t]);
length[t].display ("tree size2",fout, NULL, NULL, &length_fit[t]);
}

for ( int op = 1; (op<problem->funccount)&&(op<funccount_max); op++) //surpress NUMBER opcode
for ( int d = 1; d <= MaxDepth; d++ )
{
	problem->WriteTreeName(t, fout);
	fout << "\t" << problem->funclist[op]->getname() << "'";
	fout << d  << "\t";

	if ( counts[t][op][d].nfit.Cases() <=0 )
	  fout << counts[t][op][d].nfit.Cases() << endl;
	else
	{
//	fout << "occurs " << counts[t][op][d].length.Cases() << " times\t";
	counts[t][op][d].frequency.  display ("frequency ", fout );

	problem->WriteTreeName(t, fout);
	fout << "\t" << problem->funclist[op]->getname() << "'";
	fout << d  << "\t";

//nolonger supported
//	counts[t][op].length.  display ("length ", fout, 
//		&counts[t][op].frequency, &length[t] );
	
//	problem->WriteTreeName(t, fout);
//	fout << "\t" << problem->funclist[op]->getname() << "\t";

//	counts[t][op].children.display ("childs ", fout,  
//		&counts[t][op].frequency, &summary.children );

//	problem->WriteTreeName(t, fout);
//	fout << "\t" << problem->funclist[op]->getname() << "\t";

//	counts[t][op].fitness. display ("fitness", fout,  
//		&counts[t][op].frequency, &summary.fitness );

//	problem->WriteTreeName(t, fout);
//	fout << "\t" << problem->funclist[op]->getname() << "\t";
//
//	counts[t][op].nrank.   display ("nrank  ", fout,  
//		&counts[t][op].frequency, &summary.nrank );

//	problem->WriteTreeName(t, fout);
//	fout << "\t" << problem->funclist[op]->getname() << "\t";

	counts[t][op][d].nfit.    display ("nfit   ", fout,
		&counts[t][op][d].frequency, &summary.nfit );
	}//end if op code is used in tree
};//end for each op code
};//end for each tree

delete [] nfit;
}//end Pop::DisplayDStats

//fix later

void Pop::GatherPStats(int fcount[][NUM_TREES][MAXTREEDEPTH+1],
                       int pcount[pcount_size],
                       int tcount[NUM_TREES] ) const
//
//      Count total no of each opcode in the population. 
//      And record which trees they are in
//
{
memset(fcount,0,sizeof(int) * funccount_max * NUM_TREES*(MAXTREEDEPTH+1));
memset(pcount,0,sizeof(int) * pcount_size);
memset(tcount,0,sizeof(int) * NUM_TREES);

for ( UINT i = 0; i < popsize; i++ )
{
int f[funccount_max][NUM_TREES][MAXTREEDEPTH+1];
memset(f,0,sizeof(f));
int len;
for ( int t = 0; t<NUM_TREES; t++)
#ifdef PDGP
	if(params->params[pPDGP]) {
	  tcount[t] += problem->ProbLTStat(t,0,0,pop[i]->expr);
	  for(int d = params->net_limits[t]; d < params->net_limits[t+1]; d++){
	    const int sp = 1+d-params->net_limits[t];
	    assert(sp<=MAXTREEDEPTH);
	    for(int y = 0; y < params->pdgpwidth[d]; y++) {
	      const int op = pop[i]->pdgp(d,y)->op;
	      if(op<funccount_max) f[op][t][sp]++;
	      if(op == 0) {
		cout<<"ERROR zero opcode in "<<flush;
		cout<<i<<" "<<pop[i]->birth<<" "<<pop[i]->expr<<" "<<flush;
		pop[i]->write(PRETTY_NONE,cout);
		cout<<endl;
	      }
	    }
	  };//end for each op in this net
	}
	else
#endif /*PDGP*/
      { len = pop[i]->SubLen(pop[i]->tree_starts[t]);
	tcount[t] += problem->ProbLTStat(
			       t,pop[i]->tree_starts[t],len,pop[i]->expr);
        int tree_stack [EXPRLEN];
        tree_stack[0] = -1;
	int sp = 0;
	{for ( int n = pop[i]->tree_starts[t];
              n < (pop[i]->tree_starts[t] + len);
	      n++ 				)
	{
	        while (tree_stack[sp] == 0) { //remove finished branches
		        --sp; //-O3 bug in mips4 cc?
			tree_stack[sp] = tree_stack[sp] - 1;
		}
		++sp;
		const int op = pop[i]->expr[n].op;
		tree_stack[sp] = QARGNUM(op);
		if(op<funccount_max) {
		if(sp<MAXTREEDEPTH) f[op][t][sp]++;
		else                f[op][t][MAXTREEDEPTH]++;
		if(op == 0) {
		  cout<<"ERROR zero opcode in "<<flush;
		  cout<<i<<" "<<pop[i]->birth<<" "<<pop[i]->expr<<" "<<flush;
		  pop[i]->write(PRETTY_NONE,cout);
		  cout<<endl;
		}
	        }
	}};//end for each op in this tree

      };//end for each tree

problem->ProbLPStats(f, pcount);

for(int op = 0; (op<problem->funccount)&&(op<funccount_max); op++) {
	for ( int t = 0; t<NUM_TREES; t++)
	  for ( int d = 0; d<=MAXTREEDEPTH; d++)
		fcount[op][t][d] += f[op][t][d];}

};//end for whole population
}//end Pop::GatherPStats


void Pop::DisplayPStats(ostream & fout ) const //WBL

//    Gather and display frequency statistics by tree/opcode
//
{
int t;
#ifdef PDGP
t=0;//assume first net has most layers
#else
const int MaxDepth = (params->params[pMaxDepth]!=0)? 
                         params->params[pMaxDepth] : MAXTREEDEPTH;
assert(MaxDepth<=MAXTREEDEPTH);
#endif /*PDGP*/

//int f [problem->funccount][NUM_TREES]; //Total no of each opcode by tree
int f [funccount_max][NUM_TREES][MAXTREEDEPTH+1]; //Total no of each opcode by tree
int dummy[pcount_size];
int probl[NUM_TREES];
GatherPStats(f, dummy, probl); 

for ( int op = 1; (op<(problem->funccount+1))&&(op<funccount_max); op++) //surpress NUMBER opcode
{for ( int d = (MaxDepth==0)? 0:1; d <= MaxDepth; d++ )
{
//??	if(((((op-1)*MaxDepth)+d)%16)==0) fout<<endl;
	if(op>=problem->funccount)
		fout << "\tProbl";
	else {
		if(strlen(problem->funclist[op]->getname())+2<8) fout<<"\t";
		fout << problem->funclist[op]->getname();
	}
	if(d!=0) fout << "'" << d;
}
fout << endl;
}//end display header line
//fout << endl;

for ( t = 0; t<NUM_TREES; t++)
{
problem->WriteTreeName(t, fout);
for ( int op = 1; (op<(problem->funccount+1))&&(op<funccount_max); op++) //surpress NUMBER opcode
{for ( int d = (MaxDepth==0)? 0:1; d <= MaxDepth; d++ )
{
//???	if(((((op-1)*MaxDepth)+d)%16)==0) fout<<endl;
	fout << "\t";
	if(op>=problem->funccount)
		fout << probl[t];
	else if(problem->funcbag[t]->find(op)>=0) //opcode allowed in tree
		fout << f[op][t][d];
	else if(f[op][t][d]!=0)
		fout << "ERROR " << f[op][t][d] << " should be zero!\n";
}//end for each tree depth
fout << endl;
};//end for each op code
//fout << endl;
};//end for each tree

}//end Pop::DisplayPStats


void Pop::DisplayLPStats(ostream & fout ) const //WBL
{
//    Display cases of low tree/opcode frequency in the population
//
#ifndef PDGP
const int MaxDepth = (params->params[pMaxDepth]!=0)? 
                         params->params[pMaxDepth] : MAXTREEDEPTH;
assert(MaxDepth<=MAXTREEDEPTH);
#endif /*PDGP*/

//int f [problem->funccount][NUM_TREES]; //Total no of each opcode by tree
assert (problem->funccount<=funccount_max);
int f [funccount_max][NUM_TREES][MAXTREEDEPTH+1]; //Total no of each opcode by tree
int problem_count[pcount_size];
int problem_tcount[NUM_TREES];

GatherPStats(f, problem_count, problem_tcount);

const int threshold = ( (popsize/10) > 9 )? popsize/10 : 9;

for ( int t = 0; t<NUM_TREES; t++) {
BOOL first = TRUE;
for ( int op = 0; (op<(problem->funccount+1))&&(op<funccount_max); op++) {
	if(op>=problem->funccount) {//NOT an opcode
	    if(problem_tcount[t]>=0) {//interesting result
		if( problem_tcount[t] <= threshold ) {
			if(first) {
				first = FALSE;
				problem->WriteTreeName(t, fout);
				fout <<"\t";
			}
			else {  fout <<" "; }
			fout << "Problem"<<";";
			fout << problem_tcount[t];
		};//endif <= threshold
	    }
        }//end problem_tcount
	else if(problem->funcbag[t]->find(op)>=0) {//opcode allowed in tree
	  for ( int d = (MaxDepth==0)? 0:1; d <= MaxDepth; d++ )
		if( f[op][t][d] <= threshold ) {
			if(first) {
				first = FALSE;
				problem->WriteTreeName(t, fout);
				if(d!=0){fout <<"'"<<d;} fout<<"\t";
			}
			else {  fout <<" "; }
			fout << problem->funclist[op]->getname();
			if(d!=0){fout << "'"<<d;} fout<<";";
			fout << f[op][t][d];
		};//endif <= threshold
	}
	else 
	  for ( int d = (MaxDepth==0)? 0:1; d <= MaxDepth; d++ )
	    if(f[op][t][d]!=0) {
		problem->WriteTreeName(t, fout);
		fout <<"\t";
		fout << problem->funclist[op]->getname();
		fout << " ERROR " << f[op][t][d] << " ("<<d<<")";
		fout << " should be zero!\n";
	}
};//end for each op code
if(!first) fout << endl;

};//end for each tree

BOOL first = TRUE;
for ( int o = 0; o<pcount_size; o++) {
	if( problem_count[o] <= threshold ) {
		if(first) {
			first = FALSE;
			fout <<"lowprob\t";
		}
		else {  fout <<" "; }
		fout << char('A'+o) <<";";
		fout << problem_count[o];
	};//endif <= threshold
};//end for each category
if(!first) fout << endl;

}//end Pop::DisplayLPStats


extern int CrossSubTree_t_chnglen;
extern int CrossSubTree_calls;
extern int CrossFH_calls;
extern int CrossFH_choice;
extern int CrossFH_differ;
extern int CrossFH_fitdiffer;

void Pop::DisplayCStats(ostream & fout, int gentime ) const //WBL
{
if(gencount != num_crosses_cleared_at) {
int total = 0;
for ( int t = 0; t<NUM_TREES; t++) {
	total += num_crosses[t];
	fout<< float(num_crosses[t]*100)/float(gencount-num_crosses_cleared_at)
	    << " ";
};//end for each tree

fout << "= " <<float(total*100)/float(gencount-num_crosses_cleared_at);
}
else {
  for ( int t = 0; t<NUM_TREES; t++) fout<<num_crosses[t]*100<<" ";
  fout << "= 0";
}
fout << endl;

fout<<gentime<<" CrossSubTree "<<CrossSubTree_calls;
fout<<" total length change "<<CrossSubTree_t_chnglen;
fout<<" Homologous "<<CrossFH_calls;
fout<<" active choice "<<CrossFH_choice;
fout<<" different chrome "<<CrossFH_differ;
fout<<" fitness "<<CrossFH_fitdiffer<<endl;

}//end Pop::DisplayCStats


class hashchain;

struct data {
	stats dupl; stats crossover;
};

class hashtable : public ghashtable {
  typedef hashtable* hashtableptr;
public:
  hashtable(int s): ghashtable(s) {};
  void include(data* output) const; 
//friend class hashchain;
};

class hashchain : public ghashchain { 
  typedef hashchain* hashchainptr;
  int count;
  int cross; //number with identical code produced by Xover since oldest one

  inline void init(Chrome* c) {
    chrome = c;
    count = 1;
    cross = 0;
  }
  inline int Count() const { return count; };
  inline int Cross() const { return cross; };
public:
  Chrome* chrome;  //oldest chrome with identical code
  hashchain(ghashtable* t, int code,Chrome* c): ghashchain(t, code) {init(c);};
  hashchain(ghashchain* endofchain,Chrome* c):ghashchain(endofchain){init(c);};
  inline void inc()        { count++; };
  inline void inc_cross()  { cross++; };
  void include(data* answer) /*const CC.64*/ {
    for(hashchain* h = this; h != NULL; h = hashchainptr(h->chain)) {
       answer->dupl.include     (float( h->Count() ));
       answer->crossover.include(float( h->Cross() ));
    }
  }
};
typedef hashchain* hashchainptr;


ghashtable::ghashtable(int s)
{
	size = s;
	table = new ghashchain*[size];
	memset(table,NULL,sizeof(ghashchain*)*size);
}
ghashtable::~ghashtable()
{
	for(int i=0; i<size;i++) {
		if(table[i]!=NULL) {
		  ghashchain* tmp;
		  for(ghashchain* h = table[i]->chain; h != NULL; h = tmp ) {
		    tmp = h->chain;
		    delete h;
		  }
		}
	}
	delete[] table;
}
void hashtable::include(data* answer) const
{
	for(int i=0; i<size;i++) {
		if(table[i]!=NULL) 
			(hashchainptr(table[i]))->include(answer);
	}
}


int Chrome::hashcode() const
{
int code = 1;
#ifdef PDGP
if(params->params[pPDGP]) {
  for(int ip=0; ip<params->params[pMaxExpr];ip++) {
    code = code + 1 + expr[ip].op;
    for(int a=0; a<ARGNUM();a++)
      code = code + (1 + expr[ip].lk[a])*a*101;
  code *= 99; //arbitary value should be ok for NUM_TREES=5???
  }
return code;
}//endif pPDGP
#endif /*PDGP*/
#ifdef MULTREE
for (int t = 0; t < NUM_TREES; t++ ) {
	int ip = tree_starts[t];
#else
	int ip = 0;
#endif
        int args = 1;
        while (args > 0)
        {
		code = code + 1 + expr[ip].op;
		args=args+ARGNUM()-1;
		ip++;
        }
	code *= 99; //arbitary value should be ok for NUM_TREES=5???
#ifdef MULTREE
};//endfor all trees
#endif

return code;

}//end Chrome::hashcode


BOOL Chrome::same(Chrome* other) const
{

#ifdef PDGP
  if(params->params[pPDGP]) 
    return (memcmp(expr,other->expr,sizeof(node)*params->params[pMaxExpr])==0);
#endif /*PDGP*/
#ifdef MULTREE
if( memcmp(tree_starts, other->tree_starts, sizeof(tree_starts)) !=0 )
	return FALSE;

for (int t = 0; t < NUM_TREES; t++ ) {
	int ip = tree_starts[t];
#else
	int ip = 0;
#endif
        int args = 1;
        while (args > 0)
        {
		if(expr[ip].op != other->expr[ip].op) return FALSE;
		args=args+ARGNUM()-1;
		ip++;
        }
#ifdef MULTREE
};//endfor all trees
#endif

return TRUE;

}//end same

void Pop::DisplayDupStats(ostream & fout ) const //WBL

//    Gather and display Duplicate Tree statistics
//
{
int duplicates = 0;
hashtable ht(popsize);


for (UINT i=0;i<popsize;i++) { 
	int code = pop[i]->hashcode();
	hashchain* ptr = hashchainptr(ht.chainhead(code));
	if(ptr == NULL)
		new hashchain(&ht,code,pop[i]);
	else {
		BOOL found = FALSE;
		hashchain* prev = ptr;
		while(!found && ptr != NULL) {
			found = (pop[i]->same(ptr->chrome));
			prev  = ptr;
			ptr   = hashchainptr(ptr->chain); 
		}
		if(!found) new hashchain(prev,pop[i]);
		else {
			duplicates++;
			prev->inc();
			Chrome* pc = prev->chrome;
			if(pop[i]->birth < pc->birth) {
#ifdef TRACE
#ifdef MULTREE
			   //Was rev created by crossover?
			   if(pc->mum.birth >=0 && pc->mum.xtree >=0){
				   prev->inc_cross();
//				   fout<<"\nDuplicate pop "<<i
//				       <<" created by crossover1";
//				   pop[i]->write_trace(fout);
//				   pc->write_trace(fout);
//				   pc->write(PRETTY_NONE,fout);
			   }
#endif
#endif
			   prev->chrome = pop[i]; // keep oldest
			}//endif pop[i] created before prev

			else {
#ifdef TRACE
#ifdef MULTREE
			   //Was pop[i] created by crossover?
			   if(pop[i]->mum.birth >= 0 && pop[i]->mum.xtree >=0){
				   prev->inc_cross();
//				   fout<<"\nDuplicate pop "<<i
//				       <<" created by crossover2";
//				   pop[i]->write_trace(fout);
//				   pc->write_trace(fout);
//				   pop[i]->write(PRETTY_NONE,fout);
			   }
#endif
#endif
		        }//end else pop[i] created after prev

		}
	}
}//end for all population

//fout<<duplicates;

data s;
ht.include(&s);

fout<<"different "<<s.dupl.Cases();
fout<<" duplicates "<<duplicates;
fout<<" produced by crossover "<<s.crossover.Sum();
if(duplicates!=0) 
	fout<<" ratio "<<s.crossover.Sum()/duplicates<<endl;
else
	fout<<" ratio -0\n";

fout<<"dupl\t";
s.dupl.display("",fout);

fout<<"xovr\t";
s.crossover.display("",fout);

}//end Pop::DisplayDupStats


//////////////////////////////////////////////////////////////////////
//all.cc stats

struct counts{
  stats vworse;
  stats worse;
  stats same;
  stats better;
  stats fitter;
  stats zero;
};
counts neighbours[90];

void update_fitter(const int fit, const float fitter)
{
  assert(fit<90);
  neighbours[fit].fitter.include(fitter);
}

void update_neighbours(const int fit, const int vworse, const int worse, 
		                     const int same,   const int better)
{
  assert(fit<90);
  neighbours[fit].vworse.include(vworse);
  neighbours[fit].worse.include(worse);
  neighbours[fit].same.include(same);
  neighbours[fit].better.include(better);
  neighbours[fit].zero.include( (better==0)? 1:0 );
}

void display_neighbour_stats()
{
for(int fit=0;fit<90;fit++) {
  char buf[80];
  sprintf(buf,"fit %2d %5d %s",fit,neighbours[fit].vworse.Cases(),"vworse"); 
  neighbours[fit].vworse.display(buf);
  sprintf(buf,"fit %2d %5d %s",fit,neighbours[fit].vworse.Cases(),"worse "); 
  neighbours[fit].worse.display(buf);
  sprintf(buf,"fit %2d %5d %s",fit,neighbours[fit].vworse.Cases(),"same  "); 
  neighbours[fit].same.display(buf);
  sprintf(buf,"fit %2d %5d %s",fit,neighbours[fit].vworse.Cases(),"better"); 
  neighbours[fit].better.display(buf);
  sprintf(buf,"fit %2d %5d %s",fit,neighbours[fit].vworse.Cases(),"fitter"); 
  neighbours[fit].fitter.display(buf);
  sprintf(buf,"fit %2d %5d %s",fit,neighbours[fit].vworse.Cases(),"nofit "); 
  neighbours[fit].zero.display(buf);
}
}

inline int Pop::Parent(const INT32 birth) const {
  const int parent = birth - (1+gencount - 2*popsize);
  //cerr<<birth<<", gencount "<<gencount<<", parent "<<parent<<endl;
  assert(parent>=0 && parent<popsize);
  return parent;
}
void Pop::DisplaySolStats(ostream & fout ) const //WBL

//    Gather and display number of children which are solutions statistics
//
{
  const int mf = max_fitness -1; //undo increment in bool.cc r1.32
  int mum_max_fit[popsize]; memset(mum_max_fit,0,popsize*sizeof(int));
  int dad_max_fit[popsize]; memset(dad_max_fit,0,popsize*sizeof(int));
  int fit = 0;
  for (UINT i=0;i<popsize;i++) { 
    const Chrome* P = pop[i];
    if(P->nfitness->fvalue >=mf) {
      fit++; //sanity check
      if(P->mum.birth >= 0) mum_max_fit[Parent(P->mum.birth)]++;
      if(P->dad.birth >= 0) dad_max_fit[Parent(P->dad.birth)]++;
    }
  }//end for all population
  fout<<"number solutions("<< mf <<") "<<fit<< " "<<flush;

  stats mum;
  stats dad;
  stats sum;
  int max = -1;
  int I   = -1;
  int mum_max = -1;
  int dad_max = -1;
  for (UINT i=0;i<popsize;i++) { 
    mum.include(mum_max_fit[i]);
    dad.include(dad_max_fit[i]);
    const int both = mum_max_fit[i]+dad_max_fit[i];
    sum.include(both);
    if(both && both > max) {max = both; I = i;}
    if(mum_max_fit[i] > mum_max) mum_max = mum_max_fit[i];
    if(dad_max_fit[i] > dad_max) dad_max = dad_max_fit[i];
  }
  if(I<0) fout<<endl;
  else {
    fout << "parent of "<<max<<" solutions. Eg "<< I+(1+gencount - 2*popsize)<<flush;
    if(oldpop[I]) {oldpop[I]->write(PRETTY_NONE,fout); fout<<endl;}
    mum.display( "Mum_of_solution", fout );
    dad.display( "Dad_of_solution", fout );
    sum.display( "sum_of_solution", fout );

    if(mum_max || dad_max) {
      int nmum_max = 0; //hmm could use stats::Cases(); also already in mum.display() etc
      int ndad_max = 0;
      stats mum_length;
      stats mum_depth;
      stats dad_length;
      stats dad_depth;
      for (UINT i=0;i<popsize;i++) { 
	int size  = -1;
	int depth = -1;
	if(mum_max_fit[i]==mum_max) {
	  nmum_max++;
	  size = oldpop[i]->tree_starts[NUM_TREES-1] +
	    oldpop[i]->SubLen(oldpop[i]->tree_starts[NUM_TREES-1]);
	  mum_length.include(size);
	  depth = oldpop[i]->DepthMax(oldpop[i]->tree_starts[NUM_TREES-1],
				      size-1);
	  mum_depth.include(depth);
	}
	if(dad_max_fit[i]==dad_max) {
	  ndad_max++;
	  if(size<0) size = oldpop[i]->tree_starts[NUM_TREES-1] +
		       oldpop[i]->SubLen(oldpop[i]->tree_starts[NUM_TREES-1]);
	  dad_length.include(size);
	  if(depth<0)
	    depth = oldpop[i]->DepthMax(oldpop[i]->tree_starts[NUM_TREES-1],
					size-1);
	  dad_depth.include(depth);
	}
      }
      fout <<"nmum_max " << nmum_max << " ndad_max "<< ndad_max<<endl;
      mum_length.display( "Mum_of_solution_length", fout );
      mum_depth.display(  "Mum_of_solution_depth",  fout );
      dad_length.display( "Dad_of_solution_length", fout );
      dad_depth.display(  "Dad_of_solution_depth",  fout );
    }
  }
}//end Pop::DisplaySolStats


